from random import shuffle
import asyncio
from werewolf_roles import Role, Werewolf, Minion, Tanner, Hunter
from werewolf_players import Player, No_player


class Game:

	def __init__(self, bot):
		self.running = False
		self.bot = bot
		self.player_list = []
		self.role_list = Role.role_set

	def set_channel(self, channel):
		self.channel = channel

	async def send(self, message):
		await self.channel.send(message)

	async def receive(self, command):
		def check(msg):
			return msg.channel == self.channel and msg.content.startswith(command)

		return await self.bot.wait_for('message', check=check)

	def setup(self):
		self.werewolf_list = []
		self.mason_list = []

	async def set_players(self):

		await self.send("Who is playing?")
		msg = await self.receive('$players')

		# use info from last round otherwise
		if not msg.content.startswith('$players last'):
			self.player_list = [await Player.make(mem, self) for mem in msg.mentions]

		# check conditions
		if not 0 <= len(self.player_list) <= 10:
			raise ValueError("Invalid number of players: " + str(len(self.player_list)))

		# send confirmation
		await self.send("Players: " + ", ".join(p.name() for p in self.player_list))

	async def set_roles(self):
		await self.send("With which roles do you want to play?")
		msg = await self.receive('$roles')

		# use info from last round otherwise
		if not msg.content.startswith('$roles last'):
			tmp_role = [Role.match(r, self) for r in msg.content.split()[1:]]

			# invalid input
			if None in tmp_role:
				raise ValueError("Invalid list of roles: " + str(tmp_role))

			self.role_list = tmp_role

		# check condition
		if not len(self.role_list) == (len(self.player_list) + 3):
			raise ValueError("Invalid number of roles: " + str(len(self.role_list)) + " with " + str(len(self.player_list)) + " players")

		# send confirmation
		await self.send("Roles: " + ", ".join(r.name() for r in self.role_list))

	def distribute_roles(self):
		shuffle(self.role_list)
		for i in range(len(self.player_list)):
			self.player_list[i].setRole(self.role_list[i])
			self.role_list[i].setPlayer(self.player_list[i])

		self.middle_card = self.role_list[-3:]
		self.active_role = sorted(self.role_list[:-3], key=lambda x: x.order)

	async def start_night(self):
		await asyncio.gather(*[p.send("The night has begun") for p in self.player_list])

	async def send_role(self):
		await asyncio.gather(*[p.send("Your role: " + p.night_role.name()) for p in self.player_list])

	async def night_phases(self):
		await asyncio.gather(*[r.phase1() for r in self.active_role])
		await asyncio.gather(*[r.phase2() for r in self.active_role])
		await asyncio.gather(*[r.phase3() for r in self.active_role])
		await asyncio.gather(*[r.phase4() for r in self.active_role])
		await asyncio.gather(*[r.phase5() for r in self.active_role])

	async def start_day(self):
		await self.send("The day has started")

	async def vote(self, options):

		# vote
		await self.receive('$vote')
		await self.send("Vote in DM")

		await asyncio.gather(*[p.cast_vote(options) for p in self.player_list])

		await self.send("Votes\n\n" + '\n'.join(str(p) + " :arrow_right: " + str(p.vote) for p in self.player_list))

	def tally(self, options):
		for o in options:
			o.tally = 0

		for p in self.player_list:
			p.vote.tally += 1

	def who_dead(self, options):

		maxi = max(o.tally for o in options)
		dead = [p for p in self.player_list if p.tally >= maxi]
		for d in dead:
			d.dead = True
			if isinstance(d.day_role.copy, Hunter):
				dead.append(d.vote)

		return dead

	def who_won(self, dead):

		no_dead = (len(dead) == 0)
		tanner_dead = any(isinstance(d.day_role.copy, Tanner) for d in dead)
		werewolf_dead = any(isinstance(d.day_role.copy, Werewolf) for d in dead)
		werewolf_in_game = any(isinstance(p.day_role.copy, Werewolf) for p in self.player_list)
		minion_dead = any(isinstance(d.day_role.copy, Minion) for d in dead)
		minion_in_game = any(isinstance(p.day_role.copy, Minion) for p in self.player_list)

		werewolf_won = False
		village_won = False
		tanner_won = False

		# could make it shorter using boolean algebra
		if no_dead:
			if werewolf_in_game:
				werewolf_won = True
			else:
				village_won = True
		else:
			if tanner_dead:
				tanner_won = True

				if werewolf_dead:
					village_won = True

			else:
				if werewolf_dead:
					village_won = True
				else:
					if minion_dead:
						if werewolf_in_game:
							werewolf_won = True
						else:
							village_won = True
					else:
						if minion_in_game:
							werewolf_won = True

		return werewolf_won, village_won, tanner_won, dead

	async def result(self, werewolf_won, village_won, tanner_won, dead):

		if werewolf_won:
			await self.send("Werewolves won!")

		if village_won:
			await self.send("Village won!")

		for d in dead:
			if isinstance(d.day_role.copy, Tanner):
				await self.send(str(d) + " won a tanner")

		await self.send(":skull: " + ', '.join(str(d) for d in dead))
		await self.send('\n'.join(":ballot_box " + str(p.tally) + " votes  for " + str(p) + " who is " + str(p.day_role) + " (was " + str(p.night_role) + ") " for p in self.player_list))
		await self.send("Middle cards: " + ', '.join(str(r) for r in self.middle_card))

		# debug
		await self.send("Success")

	def end(self):
		self.running = False

	async def game(self):

		try:

			self.setup()
			await self.set_players()
			await self.set_roles()
			self.distribute_roles()
			await self.start_night()
			await self.send_role()

			await self.night_phases()

			await self.start_day()
			# discussion timer

			options = self.player_list + [No_player(self)]
			await self.vote(options)
			self.tally(options)
			await self.result(*self.who_won(self.who_dead(options)))

		except ValueError as error:
			await self.send(error)
		except asyncio.TimeoutError:
			await self.send("Error: I got bored waiting for your input")
		finally:
			self.end()
			await self.send("Game ended")
